from live_cells import watch, mutable, batch, computed
from .util import LifecycleCounter, LifecycleTestCell, ValueTestObserver, observe

class WatchTracker:
    def __init__(self):
        self.values = []

    def add(self, value):
        self.values.append(value)

class TestCellWatcher:
    """Test cell watch functions defined with `watch`"""

    def test_init(self, test_watch):
        """Test that watch function is called on initialization."""

        a = mutable(1)
        b = mutable(2)

        track = WatchTracker()

        @test_watch
        def watch_sum():
            track.add((a(), b()))

        assert track.values == [(1,2)]

    def test_called_on_change(self, test_watch):
        """Test that the watch function is called when cell values change."""

        a = mutable(1)
        b = mutable(2)

        track = WatchTracker()

        @test_watch
        def watch_sum():
            track.add((a(), b()))

        a.value = 5
        b.value = 10

        assert track.values == [(1,2), (5,2), (5,10)]

    def test_called_during_batch(self, test_watch):
        """Test that the watch function is called correctly during a batch update."""

        a = mutable(1)
        b = mutable(2)

        track = WatchTracker()

        @test_watch
        def watch_sum():
            track.add((a(), b()))

        with batch():
            a.value = 5
            b.value = 10

        assert track.values == [(1,2), (5,10)]

    def test_tracks_conditional_arguments(self, test_watch):
        """Test that cells referenced conditionally are tracked correctly."""

        a = mutable(1)
        b = mutable(10)
        select = mutable(True)

        track = WatchTracker()

        @test_watch
        def watcher():
            if select():
                track.add(a())

            else:
                track.add(b())

        a.value = 2
        select.value = False
        b.value = 5

        assert track.values == [1, 2, 10, 5]

    def test_stop(self, test_watch):
        """Test that a watch function is stopped when stop() is called."""

        a = mutable(1)
        b = mutable(2)

        track = WatchTracker()

        @test_watch
        def watcher():
            track.add((a(), b()))

        a.value = 5
        b.value = 10

        watcher.stop()

        b.value = 100
        a.value = 30

        assert track.values == [(1,2), (5,2), (5,10)]

    def test_cell_state_init(self, test_watch):
        """Test that init is called on the watched cell's state."""

        counter = LifecycleCounter()
        cell = LifecycleTestCell(1, counter)

        track = WatchTracker()

        @test_watch
        def watcher():
            track.add(cell())

        assert counter.count_init == 1
        assert counter.count_dispose == 0

    def test_cell_state_dispose(self, test_watch):
        """Test that dispose is called on the watched cell's state when watcher is stopped."""

        counter = LifecycleCounter()
        cell = LifecycleTestCell(1, counter)

        track = WatchTracker()

        @test_watch
        def watcher():
            track.add(cell())

        watcher.stop()

        assert counter.count_init == 1
        assert counter.count_dispose == 1


    def test_cell_state_dispose_not_called(self, test_watch):
        """Test that dispose is not called when not all watchers have been stopped."""

        counter = LifecycleCounter()
        cell = LifecycleTestCell(1, counter)

        track = WatchTracker()

        @test_watch
        def watcher1():
            track.add(cell())

        @test_watch
        def watcher2():
            track.add(cell())

        watcher1.stop()

        assert counter.count_init == 1
        assert counter.count_dispose == 0

    def test_schedule(self, test_watch):
        """Test watch function scheduling."""

        a = mutable(0)
        b = mutable(1)

        calls = []
        first = True

        def schedule(f):
            nonlocal first

            if first:
                first = False
                f()

            else:
                calls.append(f)

        @test_watch(schedule = schedule)
        def watch_ab():
            return (a(), b())

        a.value = 5
        b.value = 10

        a.value = 23
        b.value = 33

        assert len(calls) == 4

        assert calls[0]() == (5, 1)
        assert calls[1]() == (5, 10)
        assert calls[2]() == (23, 10)
        assert calls[3]() == (23, 33)


class TestChangesOnly:
    """Test the changesOnly option"""

    def test_watch_not_called_when_value_unchanged(self, test_watch):
        """Test that the watch function is not called when the value has not changed."""

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)

        track = WatchTracker()

        @test_watch
        def watch_b():
            track.add(b())

        a.value = [4, 2, 6]

        assert track.values == [2]

    def test_watch_not_called_when_value_unchanged_in_batch(self, test_watch):
        """Test that the watch function is not called when the value has not changed with batch()."""

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)

        track = WatchTracker()

        @test_watch
        def watch_b():
            track.add(b())

        with batch():
            a.value = [4, 2, 6]

        assert track.values == [2]

    def test_watch_called_when_value_changed_after_not_changing(self, test_watch):
        """Test that the watch function is called when the value has changed.

        This tests value changes that happen after a value change
        which did not result in the value of the cell actually
        changing.

        """

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)

        track = WatchTracker()

        @test_watch
        def watch_b():
            track.add(b())

        a.value = [4, 2, 6]
        a.value = [7, 8, 9]

        assert track.values == [2, 8]

    def test_watch_called_when_one_argument_changes(self, test_watch):
        """Test that the watch function is called when at least one argument changes."""

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)

        c = mutable(3)

        track = WatchTracker()

        @test_watch
        def watch_b_c():
            track.add((b(), c()))

        with batch():
            a.value = [4, 2, 6]
            c.value = 5

        assert track.values == [(2, 3), (2, 5)]

    def test_computed_cell_not_recomputed_when_arguments_not_changed(self, test_watch):
        """Test that the value of a computed cell is not recomputed when none of the arguments change."""

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)
        c = computed(lambda: b() * 10)

        track = WatchTracker()

        @test_watch
        def watch_c():
            track.add(c())

        a.value = [4, 2, 6]

        assert track.values == [20]

    def test_computed_cell_not_recomputed_when_arguments_not_changed_in_batch(self, test_watch):
        """Test that the value of a computed cell is not recomputed when none of the arguments change in batch()."""

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)
        c = computed(lambda: b() * 10)

        track = WatchTracker()

        @test_watch
        def watch_c():
            track.add(c())

        with batch():
            a.value = [4, 2, 6]

        assert track.values == [20]

    def test_computed_cell_recomputed_when_arguments_change(self, test_watch):
        """Test that the value of a computed cell is recomputed when the arguments change.

        This tests that the computed cell is recomputed when the
        values of the arguments change AFTER they having not changed
        during the previous update cycle.

        """

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)
        c = computed(lambda: b() * 10)

        track = WatchTracker()

        @test_watch
        def watch_c():
            track.add(c())

        a.value = [4, 2, 6]
        a.value = [7, 8, 9]

        assert track.values == [20, 80]

    def test_computed_cell_recomputed_when_one_argument_changes(self, test_watch):
        """Test that the value of a computed cell is recomputed when the only one argument changes."""

        a = mutable([1, 2, 3])
        b = computed(lambda: a()[1], changes_only=True)

        c = mutable(3)
        d = computed(lambda: b() * c())

        track = WatchTracker()

        @test_watch
        def watch_d():
            track.add(d())

        with batch():
            a.value = [4, 2, 6]
            c.value = 5

        assert track.values == [6, 10]

class TestPeekCell:
    """Test peeking cell values."""

    def test_peek_value_equals_cell_value(self):
        """Test that the value of a peek cell equals the value of the original cell."""

        cell = mutable(0)
        peek = cell.peek

        assert peek.value == 0

        cell.value = 2
        assert peek.value == 2

    def test_peek_does_not_notify_observers(self):
        """Test that peek cells do not notify their observers."""

        a = mutable(0)
        b = mutable(1)

        sum = computed(lambda: a.peek() + b())

        with observe(sum, ValueTestObserver()) as observer:
            a.value = 1
            a.value = 2
            a.value = 3
            b.value = 5
            b.value = 10
            a.value = 2
            b.value = 13

            assert observer.values == [8, 13, 15]

    def test_peek_cells_compare_equal_same_cell(self):
        """Test that peeks cells compare equal when their argument cells are equal."""

        a = mutable(0)

        p1 = a.peek
        p2 = a.peek

        assert p1 == p2
        assert hash(p1) == hash(p2)

    def test_peek_cells_compare_not_equal_different_cells(self):
        """Test that peek cells compare not equal when their argument cells are not equal."""

        a = mutable(0)
        b = mutable(0)

        p1 = a.peek
        p2 = b.peek

        assert p1 != p2
        assert p1 == p1

    def test_peek_cells_manage_same_observers(self):
        """Test that peeks cells manage the same set of observers for a given argument cell."""

        counter = LifecycleCounter()
        a = LifecycleTestCell(1, counter)

        def p():
            return a.peek

        assert counter.count_dispose == 0

        with observe(p()) as observer:
            p().remove_observer(observer)

            assert counter.count_init == 1
            assert counter.count_dispose == 1

    def test_remove_peek_observer(self):
        """Test that the correct observer is removed when removing a peek cell observer."""

        counter = LifecycleCounter()
        a = LifecycleTestCell(1, counter)

        def p():
            return a.peek

        assert counter.count_dispose == 0

        with observe(p()):
            observer = ValueTestObserver()
            p().remove_observer(observer)

            assert counter.count_init == 1
            assert counter.count_dispose == 0



import pytest

from live_cells import value

def test_constant_cell_int_value():
    """Test that the value of the cell is equal to the value given to the constructor."""

    c = value(10)
    assert c.value == 10

def test_constant_cell_string_value():
    """Test that the value of the cell is equal to the value given to the constructor."""

    c = value('Hello World')
    assert c.value == 'Hello World'

def test_compares_equal():
    """Test that constant cells compare == when holding the same value."""

    a = value(1)
    b = value(1)

    assert a == b
    assert hash(a) == hash(b)

def test_compares_not_equal():
    """Test that constant cells compare != when holding different values."""

    a = value(1)
    b = value(2)

    assert a != b
    assert a == a



from .extension import cell_function, cell_extension
from .computed import computed, none
from .keys import ValueKey

@cell_extension
def on_error(self, other, type=None):
    """Create a cell that evaluates to another cell when this cell raises an exception.

    When an exception is raised while computing the value of this
    cell, the returned cell evaluates to the value of ``other``.

    If ``type`` is not *None*, the returned cell evaluates to the value of
    ``other`` only when an exception of the same type as that provided
    in ``type`` is raised.

    .. note:

       Cells returned by this method compare equal when called on the
       same object and the same arguments are provided.

    :param other: The cell to evaluate to when this cell raises an exception.
    :type other: Cell

    :param type: Type of exceptions to handle or *None* (default) to
                 handle all exceptions.

    :returns: A new cell.
    :rtype: Cell

    """

    @computed(key = OnErrorCellKey(self, other, type))
    def cell():
        try:
            return self()

        except Exception as e:
            if type is None or isinstance(e, type):
                return other()

            raise

    return cell

@cell_extension
def error(self, all=False, type=None):
    """Create a cell that captures exceptions raised by this cell.

    The returned cell evaluates to the last exception raised while
    computing the value of this cell..

    If ``type`` is not None, the returned cell only captures
    exceptions which are of the type provided in ``type``.

    .. note:

       Cells returned by this method compare equal when called on the
       same object and the same arguments are provided.

    :param all: If *True*, the returned cell evaluates to *None* when
                this cell does not raise an exception. Defaults to
                *False*.

    :param type: The type of exceptions to capture or *None* (default)
                 to capture all exceptions.

    :returns: A new cell.
    :rtype: Cell

    .. note::

       The value of the returned cell is *None* until an exception is
       raised while computing the value of this cell, regardless of
       the value of ``all``.

    """

    @computed(key = ErrorCellKey(self, all, type), changes_only=True)
    def cell():
        try:
            self()

        except Exception as e:
            if type is None or isinstance(e, type):
                return e

        if not all:
            none()

        return None

    return cell


## Keys

class OnErrorCellKey(ValueKey):
    """Key identifying a cell created with on_error."""

class ErrorCellKey(ValueKey):
    """Key identifying a cell created with error."""



import pytest

from math import isclose, trunc, floor, ceil
from live_cells import value, mutable

class Wrapper:
    """Wraps a value provides definitions for certain numeric operators."""

    def __init__(self, value):
        self.value = value

    def __matmul__(self, other):
        return Wrapper(self.value * other.value)

    def __pos__(self):
        return Wrapper(self.value + 5)

class TestNumericExtension:
    """Tests numeric operators applied on cells."""

    def test_lt(self):
        a = value(3)
        b = mutable(8)

        c1 = a < b
        c2 = b < a

        assert c1.value
        assert not c2.value

        b.value = 1

        assert not c1.value
        assert c2.value

    def test_le(self):
        a = value(3)
        b = mutable(8)

        c1 = a <= b
        c2 = b <= a

        assert c1.value
        assert not c2.value

        b.value = 3

        assert c1.value
        assert c2.value

    def test_gt(self):
        a = value(3)
        b = mutable(8)

        c1 = a > b
        c2 = b > a

        assert not c1.value
        assert c2.value

        b.value = 1

        assert c1.value
        assert not c2.value

    def test_ge(self):
        a = value(3)
        b = mutable(8)

        c1 = a >= b
        c2 = b >= a

        assert not c1.value
        assert c2.value

        b.value = 3

        assert c1.value
        assert c2.value

    def test_add(self):
        a = value(5)
        b = mutable(6)

        c = a + b

        assert c.value == 11

        b.value = 10
        assert c.value == 15

    def test_sub(self):
        a = mutable(13)
        b = value(20)

        c = a - b

        assert c.value == -7

        a.value = 50
        assert c.value == 30

    def test_mul(self):
        a = mutable(10)
        b = value(8)

        c = a * b

        assert c.value == 80

        a.value = 20
        assert c.value == 160

    def test_matmul(self):
        a = mutable(Wrapper(10))
        b = value(Wrapper(8))

        c = a @ b

        assert c.value.value == 80

        a.value = Wrapper(20)
        assert c.value.value == 160

    def test_truediv(self):
        a = mutable(7)
        b = value(2)

        c = a / b

        assert isclose(c.value, 3.5)

        a.value = 10
        assert c.value == 5

    def test_floordiv(self):
        a = mutable(7)
        b = value(2)

        c = a // b

        assert c.value == 3

        a.value = 11
        assert c.value == 5

    def test_mod(self):
        a = mutable(17)
        b = value(3)

        c = a % b

        assert c.value == 2

        a.value = 21
        assert c.value == 0

    def test_divmod(self):
        a = mutable(17)
        b = value(3)

        c = divmod(a, b)

        assert c.value == (5, 2)

        a.value = 21
        assert c.value == (7, 0)

    def test_pow(self):
        a = mutable(3)
        b = value(3)

        c = a ** b

        assert c.value == 27

        a.value = 4
        assert c.value == 64

    def test_lshift(self):
        a = mutable(21)
        b = value(2)

        c = a << b

        assert c.value == 84

        a.value = 43
        assert c.value == 172

    def test_rshift(self):
        a = mutable(21)
        b = value(2)

        c = a >> b

        assert c.value == 5

        a.value = 43
        assert c.value == 10

    def test_and(self):
        a = mutable(19)
        b = value(7)

        c = a & b

        assert c.value == 3

        a.value = 21
        assert c.value == 5

    def test_or(self):
        a = mutable(19)
        b = value(7)

        c = a | b

        assert c.value == 23

        a.value = 11
        assert c.value == 15

    def test_xor(self):
        a = mutable(19)
        b = value(7)

        c = a ^ b

        assert c.value == 20

        a.value = 11
        assert c.value == 12

    def test_neg(self):
        a = mutable(10)
        b = -a

        assert b.value == -10

        a.value = -19
        assert b.value == 19

    def test_pos(self):
        a = mutable(Wrapper(10))
        b = +a

        assert b.value.value == 15

        a.value = Wrapper(-20)
        assert b.value.value == -15

    def test_abs(self):
        a = mutable(-10)
        b = abs(a)

        assert b.value == 10

        a.value = 15
        assert b.value == 15

    def test_invert(self):
        a = mutable(10)
        b = ~a

        assert b.value == -11

        a.value = -19
        assert b.value == 18

    def test_round(self):
        a = mutable(10.5167)
        b = round(a)
        c = round(a, value(2))

        assert isclose(b.value, 11)
        assert isclose(c.value, 10.52)

        a.value = 2.056

        assert isclose(b.value, 2)
        assert isclose(c.value, 2.06)

    def test_trunc(self):
        a = mutable(10.78)
        b = trunc(a)

        assert isclose(b.value, 10)

        a.value = -2.56

        assert isclose(b.value, -2)

    def test_floor(self):
        a = mutable(10.78)
        b = floor(a)

        assert isclose(b.value, 10)

        a.value = -2.06

        assert isclose(b.value, -3)

    def test_ceil(self):
        a = mutable(10.078)
        b = ceil(a)

        assert isclose(b.value, 11)

        a.value = -2.6

        assert isclose(b.value, -2)



from .cell import Cell

class ConstantCell(Cell):
    """A cell holding a constant value.

    NOTE: ConstantCell's holding the same values compare equal.

    """

    def __init__(self, value):
        """Create a constant cell holding `value`."""

        self._value = value

    @property
    def value(self):
        return self._value

    def add_observer(self, observer):
        pass

    def remove_observer(self, observer):
        pass

    def __eq__(self, other):
        if isinstance(other, ConstantCell):
            return self.value == other.value

        return NotImplemented

    def __hash__(self):
        return hash(self.value)

def value(value):
    """Create a cell with a constant ``value``.

    The value of a constant cell never changes and its observers are
    never notified.

    .. note::

       Two constant cells compare equal if their values are equal.

    :param value: The constant value of the cell.

    :returns: A constant cell.

    """

    return ConstantCell(value)



import functools

from .cell import Cell
from .compute_cell import ComputeCell

def cell_function(compute=None, key=None):
    """Create a function that is applied on cells and returns a new cell.

    This transform the function `compute` into a function that accepts
    cells as arguments and returns a new ComputeCell which evaluates
    to `compute` applied on the values of the argument cells.

    If `key` is not None, it is called with the arguments provided to
    the `compute` function to create the key for the cell.

    Can be used a decorator.

    """

    if compute is None:
        def decorator(fn):
            return cell_function(fn, key)

        return decorator

    @functools.wraps(compute)
    def wrapper(*args):
        return ComputeCell(
            lambda: compute(*(arg.value for arg in args)), set(args),
            key = key(*args) if key is not None else None
        )

    return wrapper

def cell_extension(fn = None, name=None):
    """Add `fn` as a method to all cell objects.

    `fn` is added as a method to the Cell class under the given `name`
    or the name of `fn` if `name` is None.

    This function may be used as a decorator.

    """

    if fn is None:
        def decorator(fn):
            return cell_extension(fn, name)

        return decorator

    setattr(Cell, name if name is not None else fn.__name__, fn)

    return fn



import pytest

from live_cells import watch

@pytest.fixture()
def test_watch():
    """Provides a `watch` decorator that automatically stops the watch functions after the test."""

    watchers = []

    def decorator(fn=None, *args, **kwargs):
        if fn is None:
            return lambda f: decorator(f, *args, **kwargs)

        watcher = watch(fn, *args, **kwargs)
        watchers.append(watcher)

        return watcher

    yield decorator

    for watcher in watchers:
        watcher.stop()



from .extension import cell_function, cell_extension
from .computed import computed, none
from .compute_cell import ComputeCell
from .keys import ValueKey

@cell_extension
def logand(a, b):
    """Create a cell that computes the logical **and** of this cell and ``b``.

    .. note::

       Cells returned by this method compare equal when called on the
       same object and the same arguments are provided.

    :param b: The operand cell.
    :type b: Cell

    :returns: A cell which evaluates to the logical **and** of this
              cell's value and the value of ``b``.

    :rtype: Cell

    """

    return ComputeCell(
        lambda: a.value and b.value,
        {a, b},
        key = AndCellKey(a, b)
    )

@cell_extension
def logor(a, b):
    """Create a cell that computes the logical **or** of this cell and ``b``.

    .. note::

       Cells returned by this method compare equal when called on the
       same object and the same arguments are provided.

    :param b: The operand cell.
    :type b: Cell

    :returns: A cell which evaluates to the logical **or** of this
              cell's value and the value of ``b``.

    :rtype: Cell

    """

    return ComputeCell(
        lambda: a.value or b.value,
        {a, b},
        key = OrCellKey(a, b)
    )

@cell_extension
@cell_function(key=lambda a: NotCellKey(a))
def lognot(a):
    """Create a cell that computes the logical not of this cell.

    .. note::

       Cells returned by this method compare equal when called on the
       same object.

    :returns: A cell which evaluates to the logical not of the value
              of this cell.

    :rtype: Cell

    """

    return not a

@cell_extension
def select(cond, if_true, if_false=None):
    """Create a new cell which selects between the values of two cells based on ``cond``.

    If the value of ``cond`` is *True*, the cell evaluates to the
    value of the ``if_true`` cell. If ``cond`` is *False*, the cell
    evaluates to the value of the ``if_false`` cell.

    If ``if_false`` is *None*, the previous value of the cell is
    preserved when ``cond`` is *False*.

    :param cond: A condition cell evaluating to a boolean.
    :type cond: Cell

    :param if_true: Cell to evaluate to when ``cond`` is *True*.
    :type if_true: Cell

    :param if_false: Cell to evaluate to when ``cond`` is *False*. If
                     *None* (the default), the value of the returned
                     cell is preserved when ``cond`` evaluates to
                     *False*.

    :type if_false: Cell

    :returns: A new cell.
    :type: Cell

    """
    if if_false is None:
        return computed(lambda: if_true() if cond() else none())

    else:
        return computed(lambda: if_true() if cond() else if_false())

## Keys

class AndCellKey(ValueKey):
    """Key identifying a cell created with logand."""

    pass

class OrCellKey(ValueKey):
    """Key identifying a cell created with logor."""

    pass

class NotCellKey(ValueKey):
    """Key identifying a cell created with not."""

    pass



from .cell import Cell
from .constant_cell import value
from .mutable_cell import mutable, batch, batched
from .computed import computed, none
from .watch import CellWatcher, watch
from .exceptions import StopComputeException

from . import numeric, boolean, error_handling, peek_cell

__all__ = [
    'Cell',

    'value',
    'mutable',
    'batch',
    'batched',

    'computed',
    'none',

    'CellWatcher',
    'watch',

    'StopComputeException'
]



class ChangesOnlyState:
    """ComputeCellState mixin that adds checks for whether the cell's value has changed.

    This mixin implements did_change by comparing the new value of the
    cell to its previous value using `==`. Additionally, this state
    also converts a lazy cell to an eager cell since it is necessary
    to compute the value of the cell at every update.

    """

    def __init__(self, cell, key=None):
        super().__init__(
            cell = cell,
            key = key
        )

        self._has_value = False
        self._old_value = None

    def did_change(self):
        try:
            return not self._has_value or self.cell.value != self._old_value

        except:
            return True

    def pre_update(self):
        super().pre_update()

        try:
            self._has_value = True
            self._old_value = self.cell.value

        except:
            self._has_value = False

    def post_update(self):
        super().post_update()

        self._has_value = False
        self._old_value = None



from .cell import Cell
from .extension import cell_extension

@cell_extension(name='peek')
@property
def peek(self):
    """Read the value of this cell without reacting to changes.

    This property returns a cell that has the same value as this cell
    but does not notify its observers when the value of this cell
    changes.

    This should be used rather than accessing the :any:`Cell.value`
    property directly to ensure that this cell is active.

    """

    return PeekCell(self)

class PeekCell(Cell):
    """A cell that has the same value as another cell but does not notify its observers."""

    def __init__(self, cell):
        """Create a PeekCell that evaluates to the value of `cell`."""

        super().__init__()

        self._cell = cell

    @property
    def value(self):
        return self._cell.value

    def add_observer(self, observer):
        self._cell.add_observer(PeekCellObserver(observer))

    def remove_observer(self, observer):
        self._cell.remove_observer(PeekCellObserver(observer))

    def __eq__(self, other):
        if isinstance(other, PeekCell):
            return self._cell == other._cell

        return NotImplemented

    def __hash__(self):
        return hash(self._cell)

class PeekCellObserver:
    """Wrapper that prevents cell observer methods from being called."""

    def __init__(self, observer):
        """Create a wrapper for an `observer`."""

        self._observer = observer

    def update(self, arg, did_change):
        pass

    def will_update(self, arg):
        pass

    def __eq__(self, other):
        if isinstance(other, PeekCellObserver):
            return self._observer == other._observer

        return NotImplemented

    def __hash__(self):
        return hash(self._observer)



from .stateful_cell import StatefulCell

class InactivePersistentStatefulCellError(Exception):
    """Raised when attempting to access the state of a PersistentStatefulCell while it is inactive."""

    pass

class PersistentStatefulCell(StatefulCell):
    """A StatefulCell with a state that is created before the cell is observed.

    This cell guarantees that a non-disposed state is returned
    whenever `state` is accessed, even if the cell does not have any
    observers, as long as `key` is null.

    If `key` is non-null an `InactivePersistentStatefulCellError`
    exception is thrown if `state` is accessed when the cell is
    inactive.

    """

    @property
    def state(self):
        if self.key is None:
            return self._ensure_state()

        state = super().state

        if state is None:
            raise InactivePersistentStatefulCellError

        return state



from .observer_state import ObserverCellState
from .exceptions import StopComputeException

class ComputeCellState(ObserverCellState):
    """Provides a cell state that computes a value as function of one or more argument cells.

    This class provides an implementation of a `CellState` for a cell
    that computes a value as a function of one or more argument
    cells. The class takes care of observing the argument cells,
    computing the value and caching it until the argument cells
    change.

    Subclasses must implement the `compute()` method, which should
    compute and return the value of the cell.

    """

    def __init__(self, cell, key, arguments):
        """Create a computed cell state with a given set of `arguments`."""

        super().__init__(
            cell = cell,
            key = key
        )

        self._arguments = arguments
        self._has_value = False

    @property
    def arguments(self):
        """The argument cells on which the value of this cell depends."""

        return self._arguments

    def compute(self):
        """Compute the cell's value.

        Subclasses must implement this method.

        """

        raise NotImplementedError

    @property
    def value(self):
        """Retrieve the value of the cell.

        If the cached value of the cell is stale, it is recomputed by
        calling `compute()`.

        """

        if self.stale:
            try:
                self._value = self.compute()

            except StopComputeException as e:
                if not self._has_value:
                    self._value = e.default_value

            self.stale = False
            self._has_value = True

        return self._value

    @value.setter
    def value(self, v):
        """Set the value of the cell.

        NOTE: This does not inform the observers of the cell, that its
        value has changed. This method should only be used to restore
        the value of a cell from persistent storage.

        """

        self._value = v

    def init(self):
        super().init()

        for arg in self._arguments:
            arg.add_observer(self)

    def dispose(self):
        for arg in self._arguments:
            arg.remove_observer(self)

        super().dispose()



class ValueKey:
    """A cell key which is distinguished from other keys by one or more values.

    Instances of this class compare equal to other instances if they
    have the same runtime type and the `args` of both instances,
    provided during construction, compare equal.

    """

    def __init__(self, *args):
        """Create a key distinguished from other keys by the values in `args`."""

        self.args = args

    def __eq__(self, other):
        if type(self) == type(other):
            return self.args == other.args

        return NotImplemented

    def __hash__(self):
        return hash(self.args)



from .stateful_cell import CellState

class ObserverCellState(CellState):
    """Provides a cell state that observes another cell.

    This class provides an implementation of a `CellState` and the
    cell observer interface. This allows subclasses to directly add
    this state as an observer to other cells.

    The implementation of the cell observer directly forwards the
    notifications to the observers of this cell. This behaviour can be
    changed by overriding `on_will_update` and `on_update`, which are
    called when `will_update` and `update` are called, respectively.

    This class also provides the following properties which can be
    accessed by subclasses:

    - stale

      If true the cell's value should be recomputed when it is
      accessed again. Note, when the cell's value is recomputed, stale
      should be reset to false.

    - updating

      Is a cell update cycle currently in progress?

    """

    def __init__(self, cell, key=None):
        super().__init__(
            cell = cell,
            key = key
        )

        self.stale = True
        self.updating = False

        self._changed_dependencies = 0
        self._did_change = False

    def did_change(self):
        """Did the cell's value actually change during this update cycle?"""

        return True

    def pre_update(self):
        """Called before the update cycle for this cell begins."""

        pass

    def post_update(self):
        """Called after the update cycle for this cell has ended."""

        pass

    def on_will_update(self):
        """Called after `will_update()` is called.

        The base implementation calls `will_update()` on the observers of
        this cell.

        NOTE: This method is only called once per update cycle.

        """

        self.notify_will_update()

    def on_update(self, did_change):
        """Called after `update()` is called.

        The base implementation calls `update()` on the observers of
        this cell.

        NOTE: This method is only called once per update cycle.

        """

        self.notify_update(did_change=did_change)

    # Cell Observer Methods

    def will_update(self, arg):
        if not self.updating:
            assert self._changed_dependencies == 0, '''Number of changed dependencies not equal to zero at start of update cycle.

            This indicates that a bug in live_cells unless the error
            originates from a cell class provided by third party
            code.'''

            self.pre_update()

            self.updating = True
            self._did_change = False
            self._changed_dependencies = 0

            self.on_will_update()
            self.stale = True

        self._changed_dependencies += 1

    def update(self, arg, did_change):
        if self.updating:
            assert self._changed_dependencies > 0, '''More calls to .update() than .will_update()

            This indicates that a bug in live_cells unless the error
            originates from a cell class provided by third party
            code.'''

            self._changed_dependencies -= 1
            self._did_change = self._did_change or did_change

            if self._changed_dependencies == 0:
                self.stale = self.stale or self._did_change
                self.on_update(self._did_change and self.did_change())

                self.updating = False

                if self._did_change:
                    self.post_update()



import logging

from .cell import Cell

class StatefulCell(Cell):
    """Base class for a cell that maintains a state.

    The state of the cell should be stored in a separate `CellState`
    object, which is created after the cell's first observer is added
    and destroyed after its last observer is removed.

    Multiple `StatefulCell`s with the same not None key, share the
    same `CellState` object.

    """

    def __init__(self, key=None):
        """Create a StatefulCell identified by `key`.

        If `key` is None, it does not share its state with any other
        cell.

        """

        self.key = key
        self._state = None

    @property
    def state(self):
        """Retrieve the cell state if it has been created.

        If the state has not been created, or it has been destroyed,
        None is returned.

        """

        return self._get_state()

    def create_state(self):
        """Create the cell's state."""

        raise NotImplementedError

    def add_observer(self, observer):
        self._ensure_state().add_observer(observer)

    def remove_observer(self, observer):
        state = self._get_state()

        if state is not None:
            state.remove_observer(observer)

    def __eq__(self, other):
        if self is other:
            return True

        if isinstance(other, StatefulCell):
            return self.key != None and other.key != None and self.key == other.key

        return NotImplemented

    def __hash__(self):
        return hash(self.key)

    # Private

    def _ensure_state(self):
        """Retrieve the cell's state, creating it if necessary."""

        if self._state is None or self._state.disposed:
            self._state = GlobalStateMap.instance.get(self.key, self.create_state)

        return self._state

    def _get_state(self):
        """Retrieve the cell's state if it has been created."""

        if self._state is None or self._state.disposed:
            self._state = GlobalStateMap.instance.maybe_get(self.key)

        return self._state


class CellState:
    """Maintains the state of a stateful cell.

    This class is intended to serve as the base class for cell state
    classes. It provides functionality for adding, removing and
    notifying the cell's observers.

    """
    def __init__(self, cell, key=None):
        """Create a cell state object.

        The state object is created for a `cell` identified by
        `key`.

        If `key` is None then `cell` is the only cell using this state
        object.

        """

        self.cell = cell
        self.key = key

        self._observers = {}
        self._disposed = False

        self._notify_count = 0

    @property
    def disposed(self):
        """Has this state been disposed?

        Once a state has been disposed it should no longer be used.

        """

        return self._disposed

    def init(self):
        """Called before the first observer is added.

        This should be overridden by subclasses to include any setup
        logic required for the cell's functionality.

        """

        pass

    def dispose(self):
        """Called after the last observer is removed.

        This should be overridden by subclasses to include the cleanup
        logic for the cell.

        """

        self._disposed = True
        GlobalStateMap.instance.remove(self.key)

    def add_observer(self, observer):
        """Add an observer."""

        assert not self._disposed, 'CellState used after disposal.'

        if not self._observers:
            self.init()

        self._observers[observer] = self._observers.get(observer, 0) + 1

    def remove_observer(self, observer):
        """Remove an observer."""

        assert not self._disposed, 'CellState used after disposal.'

        if observer in self._observers:
            count = self._observers[observer]

            if count > 1:
                self._observers[observer] -= 1

            else:
                del self._observers[observer]

                if not self._observers:
                    self.dispose()

    def notify_will_update(self):
        """Notify the observers that the cell's value will change.

        This method calls `will_update` on each observer.

        """

        assert not self._disposed, 'CellState used after disposal.'

        self._notify_count += 1
        assert self._notify_count > 0, '''Notify count is less than or equal to zero at the start of the update cycle.

        This indicates a bug in live_cells unless the error originates
        from a cell class provided by third-party code.'''

        for observer in self._observers.copy():
            try:
                observer.will_update(self.cell)

            except:
                logging.debug('Exception raised by observer.will_update()',
                              exc_info=True)

    def notify_update(self, *, did_change=True):
        """Notify the observers that the cell's value has changed.

        This method calls `update` on each observer.

        A value of False for `did_change` indicates that the cell's
        value has not changed. A value of True indicates that it may
        have changed.

        """

        assert not self._disposed, '''Notify count is less than zero when calling CellState.notify_update().

        This indicates a bug in live_cells unless the error originates
        from a cell class provided by third-party code.'''


        self._notify_count -= 1
        assert self._notify_count >= 0

        for observer in self._observers.copy():
            try:
                observer.update(self.cell, did_change)

            except:
                logging.debug('Exception raised by observer.update()',
                              exc_info=True)

class GlobalStateMap:
    """Mas cell keys to shared cell states."""

    _instance = None

    @classmethod
    @property
    def instance(cls):
        """Retrieve the singleton instance."""

        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    def __init__(self):
        self.states = {}

    def get(self, key, create):
        """Retrieve the state for the cell identified by `key`, creating it if it exists.

        If the state for the cell identified by `key` has not been
        created, `create()` is called to create it and the resulting
        state is stored in the map.

        """

        if key is None:
            return create()

        if key not in self.states:
            self.states[key] = create()

        return self.states[key]

    def maybe_get(self, key):
        """Retrieve the state for the cell identified by `key`.

        If the state for the cell identified by `key` has not been
        created, None is returned.

        """

        return self.states.get(key)

    def remove(self, key):
        """Remove the state for the cell identified by `key` from the map."""

        if key in self.states:
            del self.states[key]



import math
import operator as op

from .extension import cell_function, cell_extension

@cell_extension
@cell_function
def __lt__(a, b):
    return op.lt(a, b)

@cell_extension
@cell_function
def __le__(a, b):
    return op.le(a, b)

@cell_extension
@cell_function
def __gt__(a, b):
    return op.gt(a, b)

@cell_extension
@cell_function
def __ge__(a, b):
    return op.ge(a, b)

@cell_extension
@cell_function
def __add__(a, b):
    return op.add(a, b)

@cell_extension
@cell_function
def __sub__(a, b):
    return op.sub(a, b)

@cell_extension
@cell_function
def __mul__(a, b):
    return op.mul(a, b)

@cell_extension
@cell_function
def __matmul__(a, b):
    return op.matmul(a, b)

@cell_extension
@cell_function
def __truediv__(a, b):
    return op.truediv(a, b)

@cell_extension
@cell_function
def __floordiv__(a, b):
    return op.floordiv(a, b)

@cell_extension
@cell_function
def __mod__(a, b):
    return op.mod(a, b)

@cell_extension
@cell_function
def __divmod__(a, b):
    return divmod(a, b)

@cell_extension
@cell_function
def __pow__(a, n):
    return op.pow(a, n)

@cell_extension
@cell_function
def __lshift__(a, n):
    return op.lshift(a, n)

@cell_extension
@cell_function
def __rshift__(a, n):
    return op.rshift(a, n)

@cell_extension
@cell_function
def __and__(a, b):
    return op.and_(a, b)

@cell_extension
@cell_function
def __xor__(a, b):
    return op.xor(a,b)

@cell_extension
@cell_function
def __or__(a, b):
    return op.or_(a, b)

@cell_extension
@cell_function
def __neg__(a):
    return op.neg(a)

@cell_extension
@cell_function
def __pos__(a):
    return op.pos(a)

@cell_extension
@cell_function
def __abs__(a):
    return abs(a)

@cell_extension
@cell_function
def __invert__(a):
    return op.invert(a)

@cell_extension
@cell_function
def __round__(a, ndigits=None):
    return round(a, ndigits)

@cell_extension
@cell_function
def __trunc__(a):
    return math.trunc(a)

@cell_extension
@cell_function
def __floor__(a):
    return math.floor(a)

@cell_extension
@cell_function
def __ceil__(a):
    return math.ceil(a)



import logging

from .tracking import ArgumentTracker
from .exceptions import StopComputeException
from .maybe import Maybe

class CellWatcher:
    """Maintains the state of a cell watch function."""

    def __init__(self, callback, schedule=None):
        """Create and start a cell watch function.

        `callback` is called once immediately, with no arguments, to
        determine the cells that are referenced by it.

        Afterwards `callback` is called whenever at least one of the
        cells referenced within it change, until `stop()` is called.

        `schedule` is an optional function which is called with the
        callback function provided as an argument. When `schedule` is
        called, it should schedule the callback function provided to
        it to be called.

        """

        self.callback = callback
        self.observer = CellWatchObserver(callback, schedule=schedule)

    def stop(self):
        """Stop the watch function from being called for future updates.

        The watch callback will not be called again after this method
        is called.

        """

        self.observer.stop()

    def __call__(self):
        """Register the watch function.

        This is useful to restart a watch function that was stopped by
        calling stop().

        """

        self.callback()

class CellWatchObserver:
    """Watch function cell observer."""

    def __init__(self, callback, schedule=None):
        """Create a watch function observer for watch function `callback`.

        `schedule` is an optional function which is called with the
        callback function provided as an argument. When `schedule` is
        called, it should schedule the callback function provided to
        it to be called.

        """

        self.callback = callback
        self.schedule = schedule

        self.arguments = set()
        self.updating = False

        self.waiting_for_change = False

        self.call_watch()

    def stop(self):
        """Remove the observer from the watched cells."""

        for arg in self.arguments:
            arg.remove_observer(self)

        self.arguments.clear()


    def track_argument(self, arg):
        """Track cell `arg` as a dependency of the watch function."""

        if arg not in self.arguments:
            arg.add_observer(self)
            self.arguments.add(arg)

    def call_watch(self):
        """Call the watch function scheduling it if necessary."""

        if self.schedule:
            self.schedule_call()

        else:
            with ArgumentTracker(self.track_argument):
                self.call_callback()

    def call_callback(self):
        """Call the watch function."""

        try:
            return self.callback()

        except StopComputeException:
            # Stop execution of watch function

            pass

        except:
            logging.debug('Unhandled exception in watch function', exc_info=True)

    def schedule_call(self):
        """Schedule a call to the watch function callback, using ``self.schedule``."""

        arg_values = {arg: Maybe.wrap(arg) for arg in self.arguments}

        def bind_args(arg):
            if arg in arg_values:
                return arg_values[arg].unwrap()

            self.track_argument(arg)
            return arg.value

        def callback():
            with ArgumentTracker(bind_args, override=True):
                return self.call_callback()

        self.schedule(callback)

    # Cell Observer Methods

    def will_update(self, cell):
        if not self.updating:
            self.updating = True
            self.waiting_for_change = False

    def update(self, cell, did_change):
        if self.updating or (did_change and self.waiting_for_change):
            self.updating = False
            self.waiting_for_change = not did_change

            if did_change:
                # TODO: Schedule after current update cycle
                self.call_watch()

def watch(callback=None, schedule=None):
    """Register a ``callback`` to run when the values of cells change.

    ``callback`` (A function of no arguments) is called once
    immediately to determine the cells referenced within it. It is
    then called again whenever the values of the referenced cells
    change.

    To stop `callback` from being called for further changes to the
    referenced cell values, call the :any:`CellWatcher.stop` method on
    the :any:`CellWatcher` object returned by this function.

    .. attention::

       Within ``callback``, Cells should only be referenced using the
       function call syntax and never by referencing the ``value``
       property directly.

    .. note::

       This function may be used as a decorator by omitting the
       ``callback`` argument, in which case the decorated function is
       used as the watch function. The watch function's
       :any:`CellWatcher` is then referenced using the name of the
       decorated function.

    :param callback: The watch function.
    :type callback: function

    :param schedule: A function which is called with the callback
                    function provided as an argument. When this
                    function is called, it should schedule the
                    callback function provided to it to be called.

    :type schedule: function, optional

    :returns: The handle to the watch function's state.
    :rtype: CellWatcher

    """

    if callback is None:
        def decorator(fn):
            return watch(fn, schedule)

        return decorator

    return CellWatcher(callback, schedule)



import threading

class ArgumentTracker:
    """Context manager that installs an argument tracker for a given context.

    The function ``tracker`` is called whenever a cell is referenced,
    with the function call syntax, within the context managed by the
    ArgumentTracker. The referenced cell is passed as an argument.

    The previous tracker function is restored when exiting the context.

    .. code-block::

       def track_arg(arg):
           print(f'Used: {arg}')

       with ArgumentTracker(track_arg):
           ....


    :param tracker: A function of one argument, called when a cell is
                    referenced.

    :type tracker: function

    :param override: If true the ``tracker`` function can also
                     override the values of the referenced cells by
                     returning a new value for the cell.

    :type override: bool

    """

    ## Function tracking argument cells
    _data = threading.local()
    _data.track = None

    @classmethod
    def track(cls, arg):
        """Track the cell `arg` as a dependency.

        If a cell dependency tracker, registered with `with_tracker`, is in
        effect it is called with `arg`.

        """

        if cls._data.track is not None:
            return cls._data.track(arg)

        return arg.value

    def __init__(self, tracker, override=False):
        def track_fn(cell):
            tracker(cell)
            return cell.value

        self._tracker = tracker if (override or tracker is None) else track_fn

    def __enter__(self):
        self._previous = ArgumentTracker._data.track
        ArgumentTracker._data.track = self._tracker

    def __exit__(self, exception_type, exception_value, exception_traceback):
        ArgumentTracker._data.track = self._previous

def with_tracker(tracker):
    """Install a cell dependency `tracker` to be in effect within the decorated function.

    `tracker` is the function that is installed as the cell dependency
    tracker. It is called when the value of a cell is accessed using
    the function call syntax, within the decorated function, and is
    passed the same arguments as were passed to the decorated
    function. The referenced argument cell is provided after the last
    positional argument, and before the first keyword argument.

    """

    def decorator(fn):
        def wrapper(*args, **kwargs):
            with ArgumentTracker(lambda a: tracker(*args, a, **kwargs)):
                return fn(*args, **kwargs)

        return wrapper

    return decorator

def without_tracker(fn):
    """Ensure that no cell dependency tracker is in effect within the decorated function `fn`."""

    def wrapper(*args, **kwargs):
        with ArgumentTracker(None):
            return fn(*args, **kwargs)

    return wrapper



from .stateful_cell import StatefulCell
from .compute_state import ComputeCellState
from .tracking import without_tracker, ArgumentTracker
from .exceptions import StopComputeException

from .changes_only_state import ChangesOnlyState

class DynamicComputeCell(StatefulCell):
    """A computed cell that determines its arguments at runtime."""

    def __init__(self, compute, key=None, changes_only=False):
        """Create a computed cell with a given `compute` function.

        `compute` is a function of no arguments that is called to
        compute the value of the cell. The cells referenced within
        `compute`, using the function call syntax, are automatically
        registered as dependencies of the computed cell.

        The cell is identified by `key` if it is not None.

        If `changes_only` is True, the cell only notifies its
        observers if its value has actually changed.

        """

        super().__init__(key=key)

        self._compute = compute
        self._changes_only = changes_only

    @property
    @without_tracker
    def value(self):
        state = self.state

        if state is None:
            try:
                return self._compute()

            except StopComputeException as e:
                return e.default_value

        return state.value

    def create_state(self):
        if self._changes_only:
            return DynamicComputeChangesOnlyCellState(self, self.key)

        return DynamicComputeCellState(self, self.key)

class DynamicComputeCellState(ComputeCellState):
    """Maintains the state of a DynamicComputeCell."""

    def __init__(self, cell, key):
        super().__init__(
            cell = cell,
            key = key,
            arguments = set()
        )

    def track_argument(self, arg):
        """Register `arg` as a dependency of this cell."""

        if arg not in self.arguments:
            arg.add_observer(self)
            self.arguments.add(arg)

    def compute(self):
        with ArgumentTracker(self.track_argument):
            return self.cell._compute()

class DynamicComputeChangesOnlyCellState(ChangesOnlyState, DynamicComputeCellState):
    """A DynamicComputeCellState that checks whether the cell value has changed.

    This state only notifies the observers of the cell, if the new
    value of the cell is not equal to the previous value.

    """

    pass



from .dependent_cell import DependentCell

class ComputeCell(DependentCell):
    """A stateless computed cell with a user provided function.

    The value of this cell is computed whenever it is accessed.

    """

    def __init__(self, compute, arguments, key=None):
        """Create a computed cell with a given `compute` function and `arguments`.

        When the value of this cell is accessed `compute` is called
        with no arguments. It should return the cell's value.

        The cell is identified by `key` if it is not None.

        """

        super().__init__(
            arguments = arguments,
            key = key
        )

        self.compute = compute

    @property
    def value(self):
        return self.compute()



from .dynamic_compute_cell import DynamicComputeCell
from .exceptions import StopComputeException

def computed(compute=None, key = None, changes_only = False):
    """Create a computed cell with dynamically determined arguments.

    A computed cell with compute function ``compute`` is
    created. ``compute`` is a function of no arguments, that is called
    to compute the value of the cell. The cells referenced within
    ``compute``, using the function call syntax, are automatically
    registered as dependencies of the cell such that when their values
    change the value of the computed cell is recomputed.

    .. note::

       This function may be used as a decorator by omitting the
       ``compute`` argument, in which case the decorated function is
       used as the compute function. The cell is then referenced using
       the name of the decorated function.

    :param compute: Function of no arguments called to compute the
                    value of the cell.

    :type compute: function

    :param key: Key identifying the cell if not *None*. Defaults to *None*.

    :param changes_only: If *True* the cell only notifies its
                         observers if its new value is not equal to
                         its previous value. Defaults to *False*.

    :returns: A computed cell.

    """

    if compute is None:
        def decorator(fn):
            return computed(fn, key, changes_only)

        return decorator

    return DynamicComputeCell(
        compute = compute,
        key = key,
        changes_only = changes_only
    )

def none(default_value=None):
    """Stop the computation of a computed cell's value.

    When this method is called inside the value computation function
    of a computed cell, the cell's value is not updated. Instead the
    cell's current value is preserved.

    If this function is called during the computation of the cell's
    initial value, the cell's initial value is set to ``default_value``.

    :param default_value: The value to assign to the cell if called
                          during the computation of its initial value.

    :raises StopComputeException: This exception is raised to signal
                                  to the enclosing computed cell to
                                  abort the computation.

    """

    raise StopComputeException(default_value)



from contextlib import contextmanager

from .stateful_cell import CellState
from .persistent_stateful_cell import PersistentStatefulCell

class MutableCell(PersistentStatefulCell):
    """A cell which can have its value property set directly.

    When the value property is set, the observers of the cell are
    notified of the change.

    """

    def __init__(self, value = None, key = None):
        """Create a mutable cell initialized to `value`.

        The cell is identified by `key` if it is not None.

        """

        super().__init__(key=key)

        self._value = value
        self._mutable_state = None

    @property
    def value(self):
        return self.state.value

    @value.setter
    def value(self, value):
        """Set the value of the cell and notify its observers."""

        self.state.value = value

    def create_state(self):
        if self._mutable_state is None or self._mutable_state.disposed:
            self._mutable_state = self.create_mutable_state(
                self._mutable_state if self.key is None else None
            )

        return self._mutable_state

    def create_mutable_state(self, old_state):
        """Create a new mutable cell state.

        `old_state` is the previous cell state or None if no state has
        been created for this cell.

        """

        return MutableCellState(
            cell = self,
            key = self.key,
            value = old_state.value if old_state is not None else self._value
        )

class MutableCellState(CellState):
    """Maintains the state of a mutable cell."""

    ## Batch Update State

    _is_batch = False
    _batched = set()

    def __init__(self, cell, key, value):
        """Create a mutable cell state with an initial `value`."""

        super().__init__(
            cell = cell,
            key = key
        )

        self._value = value

    @property
    def value(self):
        """Retrieve the cell's value."""

        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the cell and notify the observers."""

        if self.disposed:
            self._value = value
            return

        if self.value != value:
            self.notify_will_update()
            self._value = value

            if self.is_batch:
                self.add_to_batch()

            else:
                self.notify_update()

    @property
    def is_batch(self):
        """Is a batch update currently in effect?"""

        return self._is_batch

    def add_to_batch(self):
        """Adds this state to the current batch's list of modified cell states."""

        self._batched.add(self)

    @classmethod
    def _begin_batch(cls):
        """Begin a batch update."""

        cls._is_batch = True
        cls._batched.clear()

    @classmethod
    def _end_batch(cls):
        """End the current batch update and update all the affected cells.

        This calls `notify_update()` on all cell states that were
        added to the current batch with `add_to_batch()`.

        """

        assert cls._is_batch

        cls._is_batch = False

        for state in cls._batched:
            state.notify_update()

        cls._batched.clear()

def mutable(value = None, key = None):
    """Create a mutable cell with an initial ``value``.

    The value of a mutable cell can be changed by setting its
    ``value`` property.

    :param value: The initial value of the cell, defaults to *None*.

    :param key: Key identifying the cell if not *None*, defaults to *None*.

    :returns: A mutable cell.

    """

    return MutableCell(value=value, key = key)


@contextmanager
def batch():
    """Batch changes to mutable cells within a given managed context.

    When this context manager is used, the observers of the mutable
    cells that are set within the *with* block managed by the context
    manager, are only notified when exiting the context.

    This has no effect when used when a batching is already in effect
    before the context manager is created.

    .. code-block::
       :caption: Example

       with batch():
           a.value = 1
           b.value = 2

       # Observers of `a` and `b` are only notified when
       # exiting the `with` block.

    """

    batching = not MutableCellState._is_batch

    if batching:
        MutableCellState._begin_batch()

    try:
        yield

    finally:
        if batching:
            MutableCellState._end_batch()

def batched(fn):
    """Apply the :any:`batch` context manager to the decorated function.

    This decorator is equivalent to replacing wrapping a call to
    ``fn`` with the following:

    .. code-block::

       with batch():
           fn()

    :param fn: A function
    :type fn: function

    :returns: A function that calls ``fn`` while applying cell batching.
    :rtype: function

    """

    def wrapper(*args, **kwargs):
        with batch():
            fn(*args, **kwargs)

    return wrapper



class Maybe:
    """Wraps the result or error of a cell computation.

    A Maybe can either represent the value of a cell that was computed
    successfully, in which case ``value`` is not *None* and holds the
    value, or an exception that was raised while computing the value
    of a cell, in which case ``error`` is not *None* and holds the
    raised exception.

    :param value: The computed value or *None* if an exception was
                  raised.

    :param error: The exception raised while computing the value or
                  *None* if no exception was raised.

    """

    def __init__(self, value=None, error=None):
        self._value = value
        self._error = error

    @property
    def value(self):
        """The computed value of the cell.

        This is *None* if an exception was raised while computing the
        cell's value.

        """

        return self._value

    @property
    def error(self):
        """The exception raised while computing the value of the cell.

        This is *None* if no exception was raised while computing the
        cell's value.

        """

        return self._error

    @classmethod
    def wrap(cls, compute):
        """Wrap the result of a computation in a Maybe representing a value/error.

        ``compute`` is called to compute a value. If ``compute``
        returns a value, a Maybe representing a value is returned. If
        ``compute`` raises an exception, a Maybe representing an error
        is returned.

        :param compute: A function of no arguments.
        :type compute: function

        :returns: A Maybe representing the result returned by
                  ``compute``.

        :rtype: Maybe

        """

        try:
            return cls(value = compute())

        except Exception as e:
            return cls(error = e)

    def unwrap(self):
        """Return the value or raise the exception held in this Maybe.

        :returns: The value held in this Maybe if it represents a
                  value.

        :raises: The exception held in this Maybe if it represents an
                 error.

        """

        if self.error is not None:
            raise self.error

        return self.value

    def __eq__(self, other):
        if isinstance(other, Maybe):
            return value == other.value and error == other.error

        return NotImplemented

    def __hash__(self):
        return hash((value, error))

    def __repr__(self):
        if self.error is not None:
            return f'Maybe(error={self.error})'

        else:
            return f'Maybe(value={self.value})'



from .cell import Cell

class DependentCell(Cell):
    """A cell with a value that is dependent on a set of argument cells.

    This is not a complete Cell subclass, notably the definition of
    the `value` property is missing. The definition of the `value`
    property should be defined by subclasses.

    This cell should not hold any mutable state. Instead the `value`
    property should compute the cells value whenever it is accessed.

    """

    def __init__(self, arguments, key=None):
        """Create a cell dependent on the cells in `arguments`.

        The cell is identified by `key` if it is not None.

        """

        self._key = key
        self._arguments = arguments

    @property
    def key(self):
        """Key identifying the cell."""

        return self._key

    @property
    def arguments(self):
        """Set of cells on which the value of this cell depends."""

        return self._arguments

    def add_observer(self, observer):
        for arg in self.arguments:
            arg.add_observer(
                ObserverWrapper(observer)
            )

    def remove_observer(self, observer):
        for arg in self.arguments:
            arg.remove_observer(
                ObserverWrapper(observer)
            )

    def __eq__(self, other):
        if self is other:
            return True

        if isinstance(other, DependentCell):
            return self.key != None and other.key != None and self.key == other.key

        return NotImplemented

    def __hash__(self):
        return hash(self.key)

class ObserverWrapper:
    """Wrapper that replaces the cell instance passed to a cell observer."""

    def __init__(self, cell, observer):
        """Create a wrapper replacing the cell passed to observer with `cell`."""

        self.cell = cell
        self.observer = observer

    def update(self, arg, did_change):
        self.observer.update(self.cell, did_change)

    def will_update(self, arg):
        self.observer.will_update(self.cell)

    def __eq__(self, other):
        if isinstance(other, ObserverWrapper):
            return self.observer == other.observer and self.cell == other.cell

        return NotImplemented

    def __hash__(self):
        return hash(self.observer);



from live_cells import mutable, batch, batched, computed
from live_cells.stateful_cell import GlobalStateMap

from .util import observe, CountTestObserver, ValueTestObserver

def test_initial_value():
    """Test that the initial value of a mutable cell is the value given in constructor."""

    cell = mutable(15)
    assert cell.value == 15

def test_set_value():
    """Test that setting the value changes the cell's value."""

    cell = mutable(15)

    cell.value = 23
    cell.value = 101

    assert cell.value == 101

def test_notifies_observer():
    """Test that setting the value notifies the observer."""

    cell = mutable(15)
    observer = CountTestObserver()

    with observe(cell, observer):
        cell.value = 23
        assert observer.count_update == 1

def test_notifies_observers_twice():
    """Test that setting value twice notifies observers twice."""

    cell = mutable(15)
    observer = CountTestObserver()

    with observe(cell, observer):
        cell.value = 23
        cell.value = 101

        assert observer.count_update == 2

def test_remove_observer():
    """Test that observer not notified after it is removed."""

    cell = mutable(15)
    observer = CountTestObserver()

    with observe(cell, observer):
        cell.value = 23

    cell.value = 101

    assert observer.count_update == 1

def test_new_value_equals_old_value():
    """Test that observers are not called when the new value equals old value."""

    cell = mutable(56)
    observer = CountTestObserver()

    with observe(cell, observer):
        cell.value = 56

        assert observer.count_update == 0

def test_all_observers_notified():
    """Test that all observers are notified."""

    cell = mutable(3)

    observer1 = CountTestObserver()
    observer2 = CountTestObserver()

    with observe(cell, observer1):
        cell.value = 5

        with observe(cell, observer2):
            cell.value = 8
            cell.value = 12

            assert observer1.count_update == 3
            assert observer2.count_update == 2

def test_observed_values():
    """Test that the current value is observed when observers are notified."""

    cell = mutable('hello')
    observer = ValueTestObserver()

    with observe(cell, observer):
        cell.value = 'hi'
        cell.value = 'bye'

        assert observer.values == ['hi', 'bye']

def test_batch_update():
    """Test that all cells are updated correctly during batch update."""

    a = mutable(0)
    b = mutable(0)
    op = mutable('')

    sum = computed(lambda: a() + b())
    msg = computed(lambda: f'{a()} {op()} {b()} = {sum()}')

    observer = ValueTestObserver()

    with observe(msg, observer):

        @batched
        def set_cells1():
            a.value = 1
            b.value = 2
            op.value = '+'

        set_cells1()

        with batch():
            a.value = 5
            b.value = 6
            op.value = 'plus'

        assert observer.values == ['1 + 2 = 3', '5 plus 6 = 11']

def test_compares_equal():
    """Test that two cells compare equal when they have the same keys."""

    a = mutable(0, key = 'mutable-cell-key1')
    b = mutable(0, key = 'mutable-cell-key1')

    assert a == b
    assert hash(a) == hash(b)

def test_compares_not_equal():
    """Test that two cells compare not equal when they have different keys."""

    a = mutable(0, key = 'mutable-cell-key1')
    b = mutable(0, key = 'mutable-cell-key2')

    assert a != b
    assert a == a


def test_compares_not_equal_none_keys():
    """Test that two cells compare not equal when they have None keys."""

    a = mutable(0)
    b = mutable(0)

    assert a != b
    assert a == a

def test_state_sharing():
    """Test that cells with the same key share the same state."""

    a = mutable(0, key = 'mutable-cell-key1')
    b = mutable(0, key = 'mutable-cell-key1')

    with observe(a):
        with observe(b):
            a.value = 10

            assert a.value == 10
            assert b.value == 10

def test_state_sharing_observers():
    """Test that cells with the same key manage the same set of observers."""

    a = mutable(0, key = 'mutable-cell-key1')
    b = mutable(0, key = 'mutable-cell-key1')

    with observe(a):
        with observe(b):
            observer1 = ValueTestObserver()
            observer2 = ValueTestObserver()

            with observe(a, observer1):
                with observe(b, observer2):
                    a.value = 10
                    b.value = 20
                    a.value = 30
                    b.value = 40

                    b.remove_observer(observer1)
                    a.value = 50

                    a.remove_observer(observer2)
                    a.value = 60
                    b.value = 70

                    assert observer1.values == [10, 20, 30, 40]
                    assert observer2.values == [10, 20, 30, 40, 50]

def test_no_state_sharing():
    """Test that cells with different keys do not share the same state."""

    a = mutable(0, key = 'mutable-cell-key1')
    b = mutable(0, key = 'mutable-cell-key2')

    with observe(a):
        with observe(b):
            a.value = 10

            assert a.value == 10
            assert b.value == 0

def test_no_state_sharing_observers():
    """Test that cells with different keys do manage different sets of observers."""

    a = mutable(0, key = 'mutable-cell-key1')
    b = mutable(0, key = 'mutable-cell-key2')

    with observe(a):
        with observe(b):
            observer1 = ValueTestObserver()
            observer2 = ValueTestObserver()

            with observe(a, observer1):
                with observe(b, observer2):
                    a.value = 10
                    b.value = 20
                    a.value = 30
                    b.value = 40

                    b.remove_observer(observer1)
                    a.value = 50

                    a.remove_observer(observer2)
                    a.value = 60
                    b.value = 70

                    assert observer1.values == [10, 30, 50, 60]
                    assert observer2.values == [20, 40, 70]

def test_state_recreation():
    """Test that the cell state is recreated after disposal."""

    a = mutable(0, key='mutable-cell-key1')

    observer = CountTestObserver()
    with observe(a, observer):
        assert a.value == 0

        a.value = 10
        a.value = 15

        assert a.value == 15

    with observe(a, observer):
        assert a.value == 0

        a.value = 17

        assert a.value == 17

def test_no_leaks():
    """Test that cells with keys do not leak resources."""

    a = mutable(0, key='mutable-cell-key1')

    with observe(a):
        assert GlobalStateMap.instance.maybe_get('mutable-cell-key1') is not None

    assert GlobalStateMap.instance.maybe_get('mutable-cell-key1') is None



import pytest

from live_cells import value, mutable, computed, none
from .util import (
    observe,
    CountTestObserver, ValueTestObserver,
    MockException,
    LifecycleCounter, LifecycleTestCell
)

class TestDynamicComputedCell:
    """Tests the behaviour of a computed cell that discovers its dependencies at runtime."""

    def test_function_on_constant(self):
        """Test a basic function applied on a constant cell."""

        a = value(1)
        b = computed(lambda: a() + 1)

        assert b.value == 2

    def test_recompute_on_argument_change(self):
        """Test that computed cell is recomputed when argument values change."""

        a = mutable(1)
        b = computed(lambda: a() + 1)

        observer = ValueTestObserver()

        with observe(b, observer):
            a.value = 5
            a.value = 10

            assert observer.values == [6, 11]

    def test_recompute_when_all_argument_change(self):
        """Test that the cell is recomputed when all the arguments change."""

        a = mutable(1)
        b = mutable(1)

        c = computed(lambda: a() + b())

        observer = ValueTestObserver()

        with observe(c, observer):
            a.value = 5
            b.value = 8
            a.value = 100

            assert observer.values == [6, 13, 108]

    def test_observer_removal(self):
        """Test that the observer is not called after it is removed."""

        a = mutable(1)
        b = mutable(2)

        c = computed(lambda: a() + b())

        observer = ValueTestObserver()

        with observe(c, observer):
            a.value = 8
            b.value = 10

        a.value = 100
        b.value = 1000

        assert observer.values == [10, 18]

    def test_all_observers_notified(self):
        """Test that all observers are notified when value changes."""

        a = mutable(1)
        b = mutable(2)

        c = computed(lambda: a() + b())

        observer1 = ValueTestObserver()
        observer2 = ValueTestObserver()

        with observe(c, observer1):
            a.value = 8

            with observe(c, observer2):
                b.value = 10
                a.value = 100

                assert observer1.values == [10, 18, 110]
                assert observer2.values == [18, 110]

    def test_argument_tracking_in_conditionals(self):
        """Test that arguments are tracked correctly when appearing in conditionals."""

        a = mutable(True)
        b = mutable(2)
        c = mutable(3)

        d = computed(lambda: b() if a() else c())

        observer = ValueTestObserver()

        with observe(d, observer):
            b.value = 1
            a.value = False
            c.value = 10

            assert observer.values == [1, 3, 10]

    def test_none_preserves_previous_value(self):
        """Test that the previous value is preserved when none() is used."""

        a = mutable(0)
        evens = computed(lambda: a() if a() % 2 == 0 else none())

        observer = ValueTestObserver()

        with observe(evens, observer):
            a.value = 1
            a.value = 2
            a.value = 3
            a.value = 4
            a.value = 5

        assert observer.values == [0, 2, 4]

    def test_cell_initialized_to_none_default_value(self):
        """Test that the value of the cell is initialized to the default value given in none()."""

        a = mutable(1)
        evens = computed(lambda: a() if a() % 2 == 0 else none(0))

        observer = ValueTestObserver()

        with observe(evens, observer):
            a.value = 3
            a.value = 4
            a.value = 5
            a.value = 6

        assert observer.values == [0, 4, 6]

    def test_cell_initialize_to_none_when_no_default_value(self):
        """Test that the value of the cell is initialized to None when no default value given in none()."""

        a = mutable(1)
        evens = computed(lambda: a() if a() % 2 == 0 else none())

        observer = ValueTestObserver()

        with observe(evens, observer):
            a.value = 3
            a.value = 4
            a.value = 5
            a.value = 6

        assert observer.values == [None, 4, 6]



    def test_computed_cell_argument(self):
        """Test that computed cell argument is tracked correctly."""

        a = mutable(True)
        b = mutable(2)
        c = mutable(3)

        d = computed(lambda: b() if a() else c())
        e = mutable(0)
        f = computed(lambda: d() + e())

        observer = ValueTestObserver()

        with observe(f, observer):
            b.value = 1
            e.value = 10
            a.value = False
            c.value = 10

            assert observer.values == [1, 11, 13, 20]

    def test_exception_on_init(self):
        """Test that exceptions in initial value are reproduced on access.

        This also tests the `computed` decorator.

        """

        @computed
        def cell():
            raise MockException

        with pytest.raises(MockException):
            cell.value

    def test_exception_while_observed(self):
        """Test that exceptions are reproduced while being observed."""

        @computed
        def cell():
            raise MockException

        with observe(cell):
            with pytest.raises(MockException):
                cell.value

    def test_compares_equal(self):
        """Test that two cells with the same keys compare equal."""

        a = mutable(0)
        b = mutable(0)

        c1 = computed(lambda: a() + b(), key='theKey')
        c2 = computed(lambda: a() + b(), key='theKey')

        assert c1 == c2
        assert hash(c1) == hash(c2)

    def test_compares_not_equal(self):
        """Test that two cells with different keys compare not equal."""

        a = mutable(0)
        b = mutable(0)

        c1 = computed(lambda: a() + b(), key='theKey1')
        c2 = computed(lambda: a() + b(), key='theKey2')

        assert c1 != c2
        assert c1 == c1

    def test_compares_not_equal_none_keys(self):
        """Test that two cells with None keys compare not equal."""

        a = mutable(0)
        b = mutable(0)

        c1 = computed(lambda: a() + b())
        c2 = computed(lambda: a() + b())

        assert c1 != c2
        assert c1 == c1

    def test_compares_equal_decorator(self):
        """Test that two cells with the same keys, defined with @computed, compare equal."""

        a = mutable(0)
        b = mutable(0)

        @computed(key='theKey')
        def c1():
            return a() + b()

        @computed(key='theKey')
        def c2():
            return a() + b()

        assert c1 == c2
        assert hash(c1) == hash(c2)

    def test_compares_not_equal_decorator(self):
        """Test that two cells with different keys, defined with @computed, compare not equal."""

        a = mutable(0)
        b = mutable(0)

        @computed(key='theKey1')
        def c1():
            return a() + b()

        @computed(key='theKey2')
        def c2():
            return a() + b()

        assert c1 != c2
        assert c1 == c1

    def test_compares_not_equal_none_keys_decorator(self):
        """Test that two cells with None keys, defined with @computed, compare not equal."""

        a = mutable(0)
        b = mutable(0)

        @computed
        def c1():
            return a() + b()

        @computed
        def c2():
            return a() + b()

        assert c1 != c2
        assert c1 == c1

    def test_shared_state(self):
        """Test that cells with the same keys manage the same observers."""

        counter = LifecycleCounter()
        a = LifecycleTestCell(0, counter)

        def f():
            return computed(lambda: a() + 1, key='theKey')

        assert counter.count_init == 0

        observer = CountTestObserver()

        with observe(f(), observer):
            assert counter.count_init == 1
            assert counter.count_dispose == 0

            f().remove_observer(observer)

            assert counter.count_init == 1
            assert counter.count_dispose == 1

    def test_state_recreated(self):
        """Test that the cell's state is recreated after it is destroyed."""

        counter = LifecycleCounter()
        a = LifecycleTestCell(0, counter)

        def f():
            return computed(lambda: a() + 1, key='theKey')

        observer = CountTestObserver()

        with observe(f(), observer):
            f().remove_observer(observer)

        with observe(f()):
            assert counter.count_init == 2
            assert counter.count_dispose == 1



## Utilities used for testing

from contextlib import contextmanager

from live_cells.stateful_cell import StatefulCell, CellState

class CountTestObserver:
    """A cell observer that counts how many times its methods are called.

    `count_will_update` is the number of times `will_update` was
    called.

    `count_update` is the number of times `update` was called.

    """

    def __init__(self):
        self.count_will_update = 0
        self.count_update = 0

    def update(self, cell, did_change):
        self.count_update += 1

    def will_update(self, cell):
        self.count_will_update += 1

class ValueTestObserver:
    """A cell observer that records the values of the observed cell in a list.

    This observer records the value of the observed cell(s), in the
    list `values`, at every call to `update`.

    NOTE: Duplicate values are not recorded in `values`.

    """

    def __init__(self):
        self.values = []

        self._updating = False
        self._notify_count = 0
        self._did_change = False

    def will_update(self, cell):
        if not self._updating:
            assert self._notify_count == 0

            self._updating = True
            self._notify_count = 0
            self._did_change = False

        self._notify_count += 1

    def update(self, cell, did_change):
        if self._updating:
            assert self._notify_count > 0

            self._did_change = self._did_change or did_change

            self._notify_count -= 1

            if self._notify_count == 0:
                self._updating = False

                if self._did_change:
                    try:
                        value = cell.value

                        if not self.values or self.values[-1] != value:
                            self.values.append(value)

                    except:
                        pass

class LifecycleCounter:
    """Keeps track of how many times init() and dispose() were called.

    `count_init`: Number of times `init()` was called.
    `count_dispose`: Number of times `dispose() was called.

    """

    def __init__(self):
        self.count_init = 0
        self.count_dispose = 0

class LifecycleTestCell(StatefulCell):
    """A cell which counts how many times the state init() and dispose() methods were called."""

    def __init__(self, value, counter):
        """Create a lifecycle method counter cell.

        The cell's value is the constant `value`.

        When `init()` or `dipose()` are called, the `count_init` and
        `count_dispose` properties are incremented respectively.

        """

        super().__init__()

        self._value = value
        self._counter = counter

    @property
    def value(self):
        return self._value

    def create_state(self):
        return LifecycleTestState(
            cell = self,
            key = None,
            counter = self._counter
        )

class LifecycleTestState(CellState):
    """Maintains the state of a LifecycleTestCell."""

    def __init__(self, cell, key, counter):
        """Initialize with LifecycleCounter `counter`."""

        super().__init__(cell=cell, key=key)

        self.counter = counter

    def init(self):
        super().init()
        self.counter.count_init += 1

    def dispose(self):
        self.counter.count_dispose += 1
        super().dispose()


class MockException(Exception):
    """Used to test propagation of exceptions."""

    pass

@contextmanager
def observe(cell, observer=CountTestObserver()):
    """Add an `observer` to `cell` for the lifetime of a managed context.

    This is a context manager that adds the observer `observer` to
    `cell`, when entering the `with` block, and automatically removes
    it when exiting the context.

    ```
    observer = CountTestObserver()

    with observer(cell, CountTestObserver()):
       # observer is an observer of `cell` within the `with` block
       ...
    ```

    """

    cell.add_observer(observer)

    # Compute value to ensure cell is active
    try:
        cell.value

    except:
        pass

    try:
        yield observer

    finally:
        cell.remove_observer(observer)



class StopComputeException(Exception):
    """Exception used to indicate that the cell's value should not be computed.

    When this exception is raised inside the value computation
    function of a cell, the cell's value is not updated. Instead the
    cell's current value is preserved.

    :param default_value: The value to initialize the cell to, if this
                          exception is thrown while computing the
                          initial value of the cell.

    """

    def __init__(self, default_value=None):
        self.default_value = default_value

        super().__init__("StopComputeException() raised. If you're seeing this you most likely"
                         ' used none() outside a cell value computation function.')



import pytest

from live_cells import value, mutable

from .util import ValueTestObserver, observe

class TestBooleanExtension:
    """Tests extension functions on boolean cells."""

    def test_and(self):
        a = mutable(True)
        b = mutable(True)

        c = a.logand(b)

        assert c.value

        a.value = False
        assert not c.value

        b.value = False
        assert not c.value

        a.value = True
        assert not c.value

        b.value = True
        assert c.value

    def test_and_compares_eq_same_cells(self):
        """Test that two cells created with logand are equal if given the same arguments."""

        a = mutable(True)
        b = mutable(True)

        c1 = a.logand(b)
        c2 = a.logand(b)

        assert c1 == c2
        assert hash(c1) == hash(c2)

    def test_and_compares_neq_different_cells(self):
        """Test that cells created with logand are not equal if given different arguments."""

        a = mutable(True)
        b = mutable(True)
        c = mutable(True)

        c1 = a.logand(b)
        c2 = b.logand(c)
        c3 = b.logand(a)

        assert c1 == c1
        assert c1 != c2
        assert c1 != c3
        assert c2 != c3

    def test_or(self):
        a = mutable(True)
        b = mutable(True)

        c = a.logor(b)

        assert c.value

        a.value = False
        assert c.value

        b.value = False
        assert not c.value

        a.value = True
        assert c.value

        b.value = True
        assert c.value

    def test_or_compares_eq_same_cells(self):
        """Test that two cells created with logor are equal if given the same arguments."""

        a = mutable(True)
        b = mutable(True)

        c1 = a.logor(b)
        c2 = a.logor(b)

        assert c1 == c2
        assert hash(c1) == hash(c2)

    def test_or_compares_neq_different_cells(self):
        """Test that cells created with logor are not equal if given different arguments."""

        a = mutable(True)
        b = mutable(True)
        c = mutable(True)

        c1 = a.logor(b)
        c2 = b.logor(c)
        c3 = b.logor(a)

        assert c1 == c1
        assert c1 != c2
        assert c1 != c3
        assert c2 != c3

    def test_not(self):
        a = mutable(True)
        b = a.lognot()

        assert not b.value

        a.value = False
        assert b.value

    def test_not_compares_eq_same_cells(self):
        """Test that cells created with lognot are equal if given the same argument."""

        a = mutable(True)

        c1 = a.lognot()
        c2 = a.lognot()

        assert c1 == c2
        assert hash(c1) == hash(c2)

    def test_not_compares_neq_same_cells(self):
        """Test that cells created with lognot are not equal if given different arguments."""

        a = mutable(True)
        b = mutable(True)

        c1 = a.lognot()
        c2 = b.lognot()

        assert c1 == c1
        assert c1 != c2

    def test_select(self):
        a = value('true')
        b = mutable('false')
        cond = mutable(True)

        c = cond.select(a, b)

        observer = ValueTestObserver()

        with observe(c, observer):
            cond.value = False
            b.value = 'else'
            cond.value = True

            assert observer.values == ['false', 'else', 'true']

    def test_select_no_else(self):
        a = mutable('true')
        cond = mutable(True)

        b = cond.select(a)

        observer = ValueTestObserver()

        with observe(b, observer):
            cond.value = False
            assert b.value == 'true'

            a.value = 'then'
            assert b.value == 'true'

            cond.value = True
            assert b.value == 'then'

            a.value = 'when'
            assert b.value == 'when'



import setuptools

# Load the long_description from README.md
with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='live-cells-py',
    version='0.1.1',
    author='Alexander Gutev',
    author_email='alex.gutev@gmail.com',
    description="A reactive programming library for Python ported from Live Cells for Dart",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alex-gutev/live_cells_py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import pytest

from live_cells import value, mutable, computed
from .util import ValueTestObserver, observe, MockException

class TestErrorHandlingExtension:
    """Tests error handling cell extension methods."""

    def test_on_error(self):
        """Test on_error method without exception type argument."""

        a = mutable(1)

        @computed
        def b():
            if a() <= 0:
                raise Exception('A generic exception')

            return a()

        c = mutable(2)
        result = b.on_error(c)

        observer = ValueTestObserver()

        with observe(result, observer):
            assert result.value == 1

            a.value = 0
            c.value = 4
            a.value = 10
            c.value = 100

            assert observer.values == [2, 4, 10]

    def test_on_error_with_type(self):
        """Test on_error method with exception type argument."""

        a = mutable(1)

        @computed
        def b():
            if a() < 0:
                raise Exception('A generic exception')

            elif a() == 0:
                raise MockException

            return a()

        c = mutable(2)
        result = b.on_error(c, type=MockException)

        observer = ValueTestObserver()

        with observe(result, observer):
            assert result.value == 1

            a.value = 0
            c.value = 4
            a.value = 10
            c.value = 100

            assert observer.values == [2, 4, 10]

            with pytest.raises(Exception):
                a.value = -1
                result.value

    def test_on_error_compares_equal_same_cells_and_type(self):
        """Test that cells created with on_error compare == for the same arguments."""

        a = mutable(0)
        e1 = a.on_error(value(-1))
        e2 = a.on_error(value(-1))

        assert e1 == e2
        assert hash(e1) == hash(e2)

    def test_on_error_compares_not_equal_different_self_cell(self):
        """Test that cells created with on_error compare != when called on different self cells."""

        a = mutable(0)
        b = mutable(0)

        e1 = a.on_error(value(-1))
        e2 = b.on_error(value(-1))

        assert e1 != e2
        assert e1 == e1

    def test_on_error_compares_not_equal_different_value_cell(self):
        """Test that cells created with on_error compare != when given different value cell."""

        a = mutable(0)

        e1 = a.on_error(value(-1))
        e2 = a.on_error(value(-2))

        assert e1 != e2
        assert e1 == e1

    def test_on_error_compares_not_equal_different_type(self):
        """Test that cells created with on_error compare != when given a different exception type."""

        a = mutable(0)

        e1 = a.on_error(value(-1))
        e2 = a.on_error(value(-1), type=MockException)
        e3 = a.on_error(value(-1), type=MockException)

        assert e1 != e2
        assert e1 != e3
        assert e2 == e3

    def test_error(self):
        """Test error() method."""

        a = mutable(1)

        @computed
        def b():
            if a() < 0:
                raise MockException

            return a()

        error = b.error()

        observer = ValueTestObserver()

        with observe(error, observer):
            assert error.value is None

            a.value = 2
            assert error.value is None

            a.value = -1
            assert isinstance(error.value, MockException)

            a.value = 3
            assert isinstance(error.value, MockException)

    def test_error_all_true(self):
        """Test that error(all=True) updates to None when no error."""

        a = mutable(1)

        @computed
        def b():
            if a() < 0:
                raise MockException

            return a()

        error = b.error(all=True)

        observer = ValueTestObserver()

        with observe(error, observer):
            assert error.value is None

            a.value = 2
            assert error.value is None

            a.value = -1
            assert isinstance(error.value, MockException)

            a.value = 3
            assert error.value is None

    def test_error_with_type(self):
        """Test error() method with exception type argument."""

        a = mutable(1)

        @computed
        def b():
            if a() < 0:
                raise Exception('A generic exception')

            elif a() == 0:
                raise MockException

            return a()

        error = b.error(type=MockException)

        with observe(error, ValueTestObserver()):
            assert error.value is None

            a.value = 2
            assert error.value is None

            a.value = 0
            assert isinstance(error.value, MockException)

            a.value = 3;
            assert isinstance(error.value, MockException)

            a.value = -1
            assert isinstance(error.value, MockException)


    def test_error_with_type_all_true(self):
        """Test error(all=True) method with exception type argument."""

        a = mutable(1)

        @computed
        def b():
            if a() < 0:
                raise Exception('A generic exception')

            elif a() == 0:
                raise MockException

            return a()

        error = b.error(all=True, type=MockException)

        with observe(error, ValueTestObserver()):
            assert error.value is None

            a.value = 2
            assert error.value is None

            a.value = 0
            assert isinstance(error.value, MockException)

            a.value = -1
            assert error.value is None

    def test_error_compares_equal_with_same_cell(self):
        """Test that cells created with error() compare == for the same cell arguments."""

        a = mutable(0)

        e1 = a.error()
        e2 = a.error()

        assert e1 == e2
        assert hash(e1) == hash(e2)

    def test_error_compares_equal_with_same_cell_and_type(self):
        """Test that cells created with error() compare == for the same cell and type arguments."""

        a = mutable(0)

        e1 = a.error(type=MockException)
        e2 = a.error(type=MockException)

        assert e1 == e2
        assert hash(e1) == hash(e2)

    def test_error_compares_not_equal_with_different_cells(self):
        """Test that cells created with error() compare != when given different cell arguments."""

        a = mutable(0)
        b = mutable(0)

        e1 = a.error()
        e2 = b.error()

        assert e1 != e2
        assert e1 == e1

    def test_error_compares_not_equal_with_different_types(self):
        """Test that cells created with error() compare != when given different type arguments."""

        a = mutable(0)

        e1 = a.error()
        e2 = a.error(type=MockException)

        assert e1 != e2
        assert e1 == e1

    def test_error_compares_not_equal_with_different_arguments(self):
        """Test that cells created with error() compare != when all arguments are different."""

        a = mutable(0)

        e1 = a.error(all=True)
        e2 = a.error(all=False)
        e3 = a.error(all=True)

        assert e1 != e2
        assert e1 == e1
        assert e1 == e3



from .tracking import ArgumentTracker

class Cell:
    """Base cell class.

    All cells should inherited this class.

    """

    @property
    def value(self):
        """The value of the cell."""

        raise NotImplementedError

    def __call__(self):
        """Retrieve the value of the cell and track it as a dependency.

        This should be used, instead of the :any:`Cell.value`
        property, when the cell needs to be registered automatically
        as a dependency.

        :returns: The value of the cell.

        """

        return ArgumentTracker.track(self)

    def add_observer(self, observer):
        """Add an ``observer`` to the cell.

        ``observer`` is notified whenever the value of this cell
        changes.

        :param observer: The observer to add to the cell.

        """

        raise NotImplementedError

    def remove_observer(self, observer):
        """Remove an ``observer`` from the cell.

        After ``observer`` is removed, it is no longer notified when
        the value of the cell changes.

        .. note::

           To remove a given ``observer``, this method has to be
           called the same number of times as `add_observer` was
           called for the same `observer`.

        :param observer: The observer to remove.

        """

        raise NotImplementedError
